from typing import Protocol, Callable, List
import logging
from bencher.bench_cfg import BenchRunCfg, BenchCfg
from bencher.variables.parametrised_sweep import ParametrizedSweep
from bencher.bencher import Bench
from bencher.bench_report import BenchReport
from copy import deepcopy


class Benchable(Protocol):
    def bench(self, run_cfg: BenchRunCfg, report: BenchReport) -> BenchCfg:
        raise NotImplementedError


class BenchRunner:
    def __init__(
        self,
        bench_class=None,
        run_cfg: BenchRunCfg = BenchRunCfg(),
        publisher: Callable = None,
        report=BenchReport(),
    ) -> None:
        self.run_cfg = BenchRunner.setup_run_cfg(run_cfg)
        self.bench_fns = []
        self.publisher = publisher
        if bench_class is not None:
            self.add_bench(bench_class)
        self.results = []
        self.report = report
        self.servers = []

    @staticmethod
    def setup_run_cfg(run_cfg: BenchRunCfg = BenchRunCfg(), level: int = 1) -> BenchRunCfg:
        run_cfg_out = deepcopy(run_cfg)
        run_cfg_out.use_sample_cache = True
        run_cfg_out.only_hash_tag = True
        run_cfg_out.level = level
        return run_cfg_out

    @staticmethod
    def from_parametrized_sweep(
        class_instance: ParametrizedSweep,
        run_cfg: BenchRunCfg = BenchRunCfg(),
        report: BenchReport = BenchReport(),
    ):
        return Bench(f"bench_{class_instance.name}", class_instance, run_cfg=run_cfg, report=report)

    def add_run(self, bench_fn: Benchable) -> None:
        self.bench_fns.append(bench_fn)

    def add_bench(self, class_instance: ParametrizedSweep) -> None:
        def cb(run_cfg: BenchRunCfg, report: BenchReport) -> BenchCfg:
            bench = BenchRunner.from_parametrized_sweep(
                class_instance, run_cfg=run_cfg, report=report
            )
            return bench.plot_sweep(f"bench_{class_instance.name}")

        self.add_run(cb)

    def run(
        self,
        min_level: int = 2,
        max_level: int = 6,
        level: int = None,
        repeats: int = 1,
        run_cfg: BenchRunCfg = None,
        publish: bool = False,
        debug: bool = True,
        show=False,
        grouped=True,
    ) -> List[BenchCfg]:
        if run_cfg is None:
            run_run_cfg = deepcopy(self.run_cfg)
        else:
            run_run_cfg = BenchRunner.setup_run_cfg(run_cfg)

        if level is not None:
            min_level = level
            max_level = level
        for r in range(1, repeats + 1):
            for lvl in range(min_level, max_level + 1):
                for bch_fn in self.bench_fns:
                    run_lvl = deepcopy(run_run_cfg)
                    run_lvl.level = lvl
                    run_lvl.repeats = r
                    logging.info(f"Running {bch_fn} at level: {lvl} with repeats:{r}")
                    if grouped:
                        res = bch_fn(run_lvl, self.report)
                    else:
                        res = bch_fn(run_lvl, BenchReport())
                    if publish and self.publisher is not None:
                        res.publish(remote_callback=self.publisher, debug=debug)
                    if show:
                        self.servers.append(res.report.show())
                    self.results.append(res)
        return self.results



import bencher as bch

from bencher.example.example_categorical import example_categorical
from bencher.example.example_floats import example_floats


# from bencher.example.example_workflow import example_floats2D_workflow, example_floats3D_workflow


if __name__ == "__main__":
    run_cfg = bch.BenchRunCfg()
    bench_runner = bch.BenchRunner(run_cfg=run_cfg)

    bench_runner.add_run(example_categorical)
    bench_runner.add_run(example_floats)
    # bench_runner.add_run(example_floats2D)
    # bench_runner.add_run(example_floats3D)
    # bench_runner.add_run(example_1D_cat)
    # bench_runner.add_run(example_1D_float)
    # bench_runner.add_run(example_pareto)
    # bench_runner.add_run(example_float_cat)
    # bench_runner.add_run(example_time_event)
    # bench_runner.add_run(example_custom_sweep)
    # bench_runner.add_run(example_holosweep)
    # bench_runner.add_run(example_holosweep_tap)
    # bench_runner.add_run(optuna_rastrigin)
    # bench_runner.add_run(example_sample_cache)

    bench_runner.run(level=2, show=True, grouped=True)

    bench_runner.report.save_index()
    bench_runner.report.show()



from typing import Callable
from dataclasses import dataclass
from sortedcontainers import SortedDict
from .utils import hash_sha1

import logging
from diskcache import Cache
from concurrent.futures import Future, ProcessPoolExecutor


class Job:
    def __init__(
        self, job_id: str, function: Callable, job_args: dict, job_key=None, tag=""
    ) -> None:
        self.job_id = job_id
        self.function = function
        self.job_args = job_args
        if job_key is None:
            self.job_key = hash_sha1(tuple(SortedDict(self.job_args).items()))
        else:
            self.job_key = job_key
        # self.cache =None
        self.tag = tag


@dataclass
class JobFuture:
    res: dict
    job_id: str

    def result(self):
        return self.res


def run_job(job: Job, cache: Cache):
    # logging.info(f"starting job:{job.job_id}")
    result = job.function(**job.job_args)
    # logging.info(f"finished job:{job.job_id}")
    if cache is not None:
        cache.set(job.job_key, result, tag=job.tag)
    return result


class JobCache:
    def __init__(
        self,
        parallel: bool = False,
        overwrite: bool = True,
        cache_name: str = "fcache",
        tag_index: bool = True,
        size_limit: int = int(100e8),
        use_cache=True,
    ):
        if use_cache:
            self.cache = Cache(f"cachedir/{cache_name}", tag_index=tag_index, size_limit=size_limit)
            logging.info(f"cache dir: {self.cache.directory}")

        else:
            self.cache = None
        if parallel:
            self.executor = ProcessPoolExecutor()
        else:
            self.executor = None

        self.overwrite = overwrite

        self.call_count = 0
        self.size_limit = size_limit

        self.worker_wrapper_call_count = 0
        self.worker_fn_call_count = 0
        self.worker_cache_call_count = 0

    def add_job(self, job: Job) -> JobFuture | Future:
        self.worker_wrapper_call_count += 1

        if self.cache is not None:
            if not self.overwrite and job.job_key in self.cache:
                logging.info(f"Found job: {job.job_id} in cache, loading...")
                # logging.info(f"Found key: {job.job_key} in cache")
                self.worker_cache_call_count += 1
                return JobFuture(self.cache[job.job_key], job.job_id)

        self.worker_fn_call_count += 1

        if self.executor is not None:
            self.overwrite_msg(job, " starting parallel job...")
            return self.executor.submit(run_job, job, self.cache)
        self.overwrite_msg(job, " starting serial job...")
        return JobFuture(run_job(job, self.cache), job.job_id)

    def overwrite_msg(self, job, suffix) -> None:
        if self.overwrite:
            # logging.info(f"Overwriting key: {job.job_key}{suffix}")
            logging.info(f"{job.job_id} OVERWRITING cache{suffix}")

        else:
            # logging.info(f"No key: {job.job_key} in cache{suffix}")
            logging.info(f"{job.job_id} NOT in cache{suffix}")

    def clear_call_counts(self) -> None:
        """Clear the worker and cache call counts, to help debug and assert caching is happening properly"""
        self.worker_wrapper_call_count = 0
        self.worker_fn_call_count = 0
        self.worker_cache_call_count = 0

    def clear(self) -> None:
        if self.cache:
            self.cache.clear()

    def clear_tag(self, tag: str) -> None:
        logging.info(f"clearing the sample cache for tag: {tag}")
        removed_vals = self.cache.evict(tag)
        logging.info(f"removed: {removed_vals} items from the cache")

    def close(self) -> None:
        if self.cache:
            self.cache.close()

    def stats(self) -> str:
        logging.info(f"job calls: {self.worker_wrapper_call_count}")
        logging.info(f"cache calls: {self.worker_cache_call_count}")
        logging.info(f"worker calls: {self.worker_fn_call_count}")
        if self.cache:
            return f"cache size :{int(self.cache.volume() / 1000000)}MB / {int(self.size_limit/1000000)}MB"
        return ""


class JobFunctionCache(JobCache):
    def __init__(
        self,
        function: Callable,
        overwrite=False,
        parallel: bool = False,
        cache_name: str = "fcache",
        tag_index: bool = True,
        size_limit: int = int(100e8),
    ):
        super().__init__(
            parallel=parallel,
            cache_name=cache_name,
            tag_index=tag_index,
            size_limit=size_limit,
            overwrite=overwrite,
        )
        self.function = function

    def call(self, **kwargs) -> JobFuture | Future:
        return self.add_job(Job(self.call_count, self.function, kwargs))
